from pathlib import Path

import pandas as pd
from sidekick import import_later

from .db import db

DATABASES = Path(__file__).parent / "databases"
DATA_LOADERS = {
    "region": import_later("mundi.extra:load_region"),
    "income_group": import_later("mundi.extra:load_income_group"),
    "type": db.column_loader("type"),
    "subtype": db.column_loader("subtype"),
    "country_code": db.column_loader("country_code"),
    "short_code": db.column_loader("short_code"),
    "long_code": db.column_loader("long_code"),
    "numeric_code": db.column_loader("numeric_code"),
    "parent_id": db.column_loader("parent_id"),
    "alt_parents": db.column_loader("alt_parents"),
}


def filter_from_data(data, df):
    """
    Filter data in the first argument to present only the rows present in the
    second.
    """
    return data.loc[df.index.dropna()]


def filtering_from_data(cols=slice(None)):
    """
    Transforms a function that loads a data frame into a function that filters
    the rows of this data frame from the data in the first argument.
    """

    def decorator(fn):
        extra = None
        valid = None

        def loader(data):
            nonlocal extra, valid

            if extra is None:
                extra = fn()[cols]
                valid = set(extra.index)

            mask = [k for k in data.index if k in valid]
            out = extra.loc[mask]
            return pd.DataFrame(out, index=data.index)

        return loader

    return decorator


def load_database(name):
    """
    Load data from databases.
    """
    return pd.read_pickle(DATABASES / name)


def register(name, loader):
    """
    Register a loader function for the given column.
    """
    if isinstance(loader, str):
        loader = import_later(loader)
    DATA_LOADERS[name] = loader


def unregister(name, loader=None):
    """
    Unregister a loader function for the given column.
    """

    registered = DATA_LOADERS.get(name)
    if registered is None:
        return
    elif loader is None:
        del DATA_LOADERS[name]
    elif loader == registered:
        del DATA_LOADERS[name]



import importlib
import os
import re
import subprocess
import sys
import time
from itertools import chain
from pathlib import Path
from typing import Dict, List, Iterable

import pandas as pd
import sidekick as sk

from ..functions import regions
from ..types import PandasT

sqlite3 = sk.import_later("sqlite3")
IS_CODE_PATH = re.compile(r"[A-Z]{2}(-\w+)?")
PICKLE_EXTENSIONS = (".pkl", ".pickle", ".pkl.gz", ".pickle.gz")
SQL_EXTENSIONS = (".sql", ".sqlite")
CSV_EXTENSIONS = (".csv", ".csv.gz", ".csv.bz2")
HDF5_EXTENSIONS = (".hdf", ".hfd5")

db = sk.deferred(regions)


def find_data_path(package) -> Path:
    """
    Find the data folder for package.

    This assumes the package is installed as as symlink from the main
    repository folder.
    """
    try:
        mod = sys.modules[package]
    except KeyError:
        mod = importlib.import_module(package)
    return Path(mod.__file__).resolve().absolute().parent.parent / "data"


def find_data_sources(package) -> Dict[str, Path]:
    """
    Return a map from codes to paths to the locations to data sources for
    each region.
    """
    out = {}
    path = find_data_path(package)
    for sub in os.listdir(str(path)):
        if IS_CODE_PATH.fullmatch(sub):
            out[sub] = path / sub
    return out


def find_data_scripts(package) -> Dict[str, List[Path]]:
    """
    Return a map from codes to paths to the locations of the prepare scripts
    for the given folder.
    """
    out = {}
    for key, path in find_data_sources(package).items():
        out[key] = [
            path / p
            for p in sorted(os.listdir(str(path)))
            if p.startswith("prepare") and p.endswith(".py")
        ]
    return out


def execute_prepare_scripts(package, verbose=False) -> None:
    """
    Execute all prepare scripts in package.
    """
    for code, scripts in find_data_scripts(package).items():
        for script in scripts:
            if verbose:
                print(f'Script: "{code}/{script.name}"')
            dir = script.parent
            name = script.name
            t0 = time.time()
            out = subprocess.check_output([sys.executable, name], cwd=dir)
            if out:
                lines = out.decode("utf8").splitlines()
                print("OUT:")
                print("\n".join(f"  - {ln.rstrip()}" for ln in lines))
            dt = time.time() - t0
            print(f"Script executed in {dt:3.2} seconds.\n\n", flush=True)


def collect_processed_paths(package, kind=None) -> Dict[str, List[Path]]:
    """
    Return a map from codes to paths to the locations of the prepare data in
    each folder.
    """
    kind = kind or ""
    out = {}
    for key, path in find_data_sources(package).items():
        path = path / "processed"
        out[key] = [
            path / p
            for p in os.listdir(str(path))
            if p.startswith(kind) or p.startswith("db")
        ]
    return out


def collect_processed_data(package, kind=None) -> PandasT:
    """
    Return a list of dataframes collected from <code>/processed folders.

    Dataframes are ordered lexicographically first by name of the collected file,
    then by code.
    """
    paths = collect_processed_paths(package, kind)
    paths = list(chain(*paths.values()))
    paths.sort(key=lambda x: x.name)
    if not paths:
        raise ValueError(f"Empty list of datasets for {package}/{kind}")

    cols = None
    datasets = []
    for p in paths:
        df = read_path(p, kind)
        df.index.name = "id"
        if isinstance(df, pd.Series):
            pass
        elif cols is None:
            cols = set(df.columns)
        elif set(df.columns) != cols:
            invalid = cols.symmetric_difference(df.columns)
            raise ValueError(
                f"different columns in two paths:\n"
                f"- {p}\n"
                f"- {paths[0]}"
                f"- Columns: {invalid}"
            )

        if package != "mundi" and not df.index.isin(db.index).all():
            invalid = set(df.index) - db.index
            raise ValueError(
                f"index contain invalid mundi codes:\n"
                f"- {p}\n"
                f"- Invalid codes: {invalid}"
            )

        datasets.append(df)

    data = pd.concat(datasets).reset_index()
    if isinstance(data, pd.Series):
        data = data.drop_duplicates(keep="last")
    elif isinstance(data.columns, pd.MultiIndex):
        col = data.columns[0]
        data = data.drop_duplicates(col, keep="last")
        data = data.set_index("id")
    else:
        data = data.drop_duplicates("id", keep="last")
        data = data.set_index("id")
    return data.sort_index()


def clean_processed_data(package, kind=None, verbose=False):
    """
    Clean all files in the <code>/processed folders.
    """
    for code, paths in collect_processed_paths(package, kind=kind).items():
        for path in paths:
            os.unlink(str(path))
            print("  -", Path("/".join(path.parts[-3:])))


def read_path(path: Path, key: str = None) -> PandasT:
    """
    Read dataframe from path. Optional key is only used with certain file types
    such as sql and hdf5. Otherwise it is ignored.
    """
    name = path.name
    if endswith(name, PICKLE_EXTENSIONS):
        return fix_string_columns_bug(pd.read_pickle(path))
    else:
        raise ValueError(f'could not infer data type: "{path}"')


def save_path(data: PandasT, path: Path, key: str = None, **kwargs) -> None:
    """
    Save dataframe at given path. Optional key is only used with certain file
    types such as sql and hdf5. Otherwise it is ignored.

    Source type is inferred from extension.
    """
    name = path.name

    if endswith(name, PICKLE_EXTENSIONS):
        kwargs.setdefault("protocol", 3)
        data.to_pickle(str(path), **kwargs)

    elif endswith(name, SQL_EXTENSIONS):
        with sqlite3.connect(path) as conn:
            data.to_sql(key, conn, if_exists="replace")

    else:
        raise ValueError(f'could not infer data type: "{path}"')


def endswith(st: str, suffixes: Iterable[str]) -> bool:
    """
    Return true if string ends with one of the given suffixes.
    """
    return any(st.endswith(ext) for ext in suffixes)


# TODO: is it a bug? report it? check which versions are affected by it
def fix_string_columns_bug(df):
    """
    It seems that pandas do not load pickled dataframes with string columns
    with pd.NA values.

    It seems to work in small dataframes, but not large(ish) ones.
    """

    assert len(df) == len(set(df.index))

    if not hasattr(df.dtypes, "items"):
        return df

    columns = list(df.columns)

    for col_name, dtype in df.dtypes.items():
        if isinstance(dtype, pd.StringDtype):
            col = df.pop(col_name).astype(str)
            col = col[col != "<NA>"]
            col = col[~col.isna()]
            col = col.astype("string")
            df[col_name] = col

    return df[columns]



import pytest

from mundi import Region


class TestRegion:
    def test_regions_are_unique(self):
        assert id(Region("BR")) == id(Region("BR"))

    def test_region_basic_usage(self):
        br = Region("BR")
        assert br.id == "BR"
        assert hash(br)
        assert {br} == {br}

        with pytest.raises(AttributeError):
            br.name = "Brazil"

    def test_region_dynamic_attribute_fetch(self):
        br = Region("BR")
        assert br.name == "Brazil"
        assert br.short_code == "BR"
        assert br.long_code == "BRA"

    def test_region_fetch_children(self):
        br = Region("BR")
        regions = br.children()
        assert Region("BR-1") in regions

    def test_region_fetch_children_deep(self):
        br = Region("BR-DF")
        regions = br.children(deep=True)
        assert set(regions) == {
            Region("BR-5301"),
            Region("BR-530101"),
            Region("BR-5300108"),
            Region("BR-SUS:5301"),
        }

        regions = br.children(deep=False)
        assert set(regions) == {Region("BR-5301"), Region("BR-SUS:5301")}

    def test_parents(self):
        brb = Region("BR-5300108")
        assert brb.parent.id == "BR-530101"
        assert brb.parent.parent.id == "BR-5301"
        assert brb.parent.parent.parent.id == "BR-DF"
        assert brb.parent.parent.parent.parent.id == "BR-5"
        assert brb.parent.parent.parent.parent.parent.id == "BR"
        assert brb.parent.parent.parent.parent.parent.parent.id == "XSA"
        assert brb.parent.parent.parent.parent.parent.parent.parent.id == "XX"

    def test_region_children_in_multiple_hierarchies(self):
        europe = Region("XEU")
        asia = Region("XAS")
        russia = Region("RU")

        assert russia in asia.children()
        assert russia not in asia.children(which="primary")

        assert russia in europe.children()
        assert russia in europe.children(which="primary")



from mundi.plugin import utils


class TestPluginUtilities:
    def test_utility_functions(self):
        path = utils.find_data_path("mundi")
        assert str(path).endswith("data")
        assert path.exists()
        assert path.is_dir()

        sources = utils.find_data_sources("mundi")
        assert set(sources) == {"BR", "XX"}
        assert all(p.exists() for p in sources.values())

        scripts = utils.find_data_scripts("mundi")
        assert set(sources) == {"BR", "XX"}
        assert all(len(ps) >= 1 for ps in scripts.values())
        assert all(str(ps[0].name).startswith("prepare") == 1 for ps in scripts.values())



from functools import partial
from itertools import chain
from operator import attrgetter
from typing import Union, Sequence

import pandas as pd
import sidekick as sk
from sidekick import X

from .loader import DATA_LOADERS

Pandas = Union[pd.Series, pd.DataFrame]


@pd.api.extensions.register_dataframe_accessor("mundi")
class MundiDataFrameAcessor:
    def __new__(cls, obj):
        if "mundi" in obj:
            return obj["mundi"]
        return super().__new__(cls)

    def __init__(self, obj):
        self._data = obj

    def __getitem__(self, cols):
        if not isinstance(cols, tuple):
            cols = (cols,)

        left, right = map(list, sk.partition_at((X == ...), cols))
        frames = [self.extra(left)] if left else []
        if right:
            _, *right = right  # remove ellipsis
            frames.append(self.extend(right))
        return pd.concat(frames, axis=1)

    def select(self, **kwargs):
        """
        Select interface of mundi.
        """
        df = self._data
        (k, v), *rest = kwargs.items()
        m = mask(df, k, v)
        for k, v in rest:
            m &= mask(df, k, v)
        return df[m.fillna(False)]

    def extend(self, cols):
        """
        Extend dataframe with the given columns.
        """
        df = self._data.copy()
        frames = extend_columns(df, cols, init=[df])
        idx_dims = max(level_dims(f.columns) for f in frames)
        data = pd.concat(frames, axis=1)
        if idx_dims > 1:
            cols = (fill_idx(x, idx_dims) for x in data.columns)
            data.columns = pd.MultiIndex.from_tuples(cols)
        return data

    def extra(self, cols):
        """
        Extend dataframe with the given columns.
        """
        data = [extend(self._data, col) for col in cols]
        return pd.concat(data, axis=1)


def mask(data: Pandas, col: str, value) -> pd.Series:
    """
    Return a boolean mask with values in which df[col] == value
    """
    if col not in data:
        data = extend(data, col)
    return data[col].__eq__(value)


def extend(data: Pandas, name: str) -> Pandas:
    """
    Load additional data from the given data source.
    """
    try:
        fn = DATA_LOADERS[name]
    except KeyError:
        msg = f"Mundi does not know how to load {name!r} columns."
        raise ValueError(msg)
    return fn(data)


def extend_columns(data: pd.DataFrame, cols: Sequence[str], init=None, repeat=False):
    """
    Extend columns for given dataframe.
    Args:
        data:
            Input dataframe.
        cols:
            List of columns names to extend.
        init:
            Initial list of frames
        repeat:
            If true, repeat columns that are already present in dataframe or the
            init list.
    """
    frames = list(init or ())
    used_columns = set(chain(data.columns, *map(attrgetter("columns"), frames)))

    for col in cols:
        if col not in used_columns or repeat:
            extra = extend(data, col)
            if not isinstance(extra, pd.DataFrame):
                extra = pd.DataFrame({col: extra})
            elif extra.shape[1] == 1 and extra.columns == [col]:
                pass
            else:
                extra.columns = add_multi_index_level(col, extra.columns)
            frames.append(extra)
    return frames


def level_dims(idx):
    try:
        shape = idx.levshape
    except AttributeError:
        return 1
    else:
        return len(shape)


def fill_idx(x, level: int) -> tuple:
    if isinstance(x, tuple):
        n = len(x)
        if n == level:
            return x
        else:
            rest = ("",) * (level - n)
            return (*x, *rest)
    else:
        rest = ("",) * (level - 1)
        return (x, *rest)


def prepend_to_tuple(level, x):
    return (level, *x) if isinstance(x, tuple) else (level, x)


def prepend_to_tuples(level, seq):
    return map(partial(prepend_to_tuple, level), seq)


def add_multi_index_level(level, idx):
    return pd.MultiIndex.from_tuples(prepend_to_tuples(level, idx))



import mundi


class TestMundiFunctions:
    def test_country_code(self):
        assert mundi.country_code("BR") == "BR"
        assert mundi.country_code("br") == "BR"
        assert mundi.country_code("BRA") == "BR"
        assert mundi.country_code("bra") == "BR"
        assert mundi.country_code("076") == "BR"
        assert mundi.country_code("Brazil") == "BR"
        assert mundi.country_code("brazil") == "BR"

    def test_code_function(self):
        assert mundi.code("Brazil") == "BR"
        assert mundi.code("brazil") == "BR"
        assert mundi.code("BR") == "BR"
        assert mundi.code("br") == "BR"
        assert mundi.code("BR-DF") == "BR-DF"
        assert mundi.code("BR/Distrito Federal") == "BR-DF"
        assert mundi.code("Brazil/Distrito Federal") == "BR-DF"
        assert mundi.code("brazil/distrito federal") == "BR-DF"
        assert mundi.code("Brazil/Brasília") == "BR-5300108"



import importlib

import pandas as pd
import sidekick as sk

from .functions import regions
from .types import PandasT

db = sk.deferred(regions)


def fix_types(package, key, data: PandasT) -> PandasT:
    """
    Fix types of input dataframe using information for the given package.
    """
    mod = importlib.import_module(package)
    try:
        types = mod.DATA_TYPES
    except AttributeError:
        msg = f'module "{package}" does not define a DATA_TYPES constant.'
        raise RuntimeError(msg)

    # Filter types to the desired database
    if isinstance(types, dict):
        try:
            columns = set(mod.DATA_COLUMNS[key])
        except AttributeError:
            msg = f'module "{package}" does not define a DATA_COLUMNS constant.'
            raise RuntimeError(msg)
        types = {k: v for k, v in types.items() if k in columns}

    return data.astype(types)


def sum_children(package, key, data: PandasT, cols=None) -> PandasT:
    """
    Aggregate children by sum.
    """

    id_name = data.index.name

    for _ in range(100):
        # Extract m2m relations to data
        m2m = child_parent_m2m(data).rename({"child_id": id_name}, axis=1)

        # Fill parents with aggregate
        new = (
            pd.merge(m2m, data.reset_index(), on=id_name)
            .set_index("parent_id")
            .dropna()
            .groupby("parent_id")
            .agg("sum")
            .reset_index()
            .rename({"parent_id": id_name}, axis=1)
            .set_index(id_name)
        )
        new = new[~new.index.isin(data.index)]

        # Join and continue or stop process
        if len(new) > 0:
            data = pd.concat([data, new], verify_integrity=False)
        else:
            break

    else:
        raise RuntimeError("maximum number of iterations")

    print(data.loc["BR-1"])
    return data


def mean_children(package, key, data: PandasT, cols=None) -> PandasT:
    """
    Aggregate children
    """
    data = _aggregate_primary(data, "mean")
    return data


def max_children(package, key, data: PandasT, cols=None) -> PandasT:
    """
    Aggregate children
    """
    raise NotImplementedError


def min_children(package, key, data: PandasT, cols=None) -> PandasT:
    """
    Aggregate children
    """
    raise NotImplementedError


def fill_from_children_population_weighted_mean(
    package, key, data: PandasT, cols=None
) -> PandasT:
    """
    Fill children
    """
    raise NotImplementedError


def as_parent(package, key, data: PandasT, cols=None) -> PandasT:
    """
    Fill missing data using the same value as the parent.
    """


def _aggregate_primary(data, agg, maxiter=500) -> PandasT:
    """
    Worker function for sum_children, mean_children, etc.
    """
    parents = db.mundi["parent_id"]
    id_name = data.index.name

    for _ in range(maxiter):
        # Join parents and data and groupby
        join = pd.concat([parents, data], axis=1)
        df = (
            join.dropna(subset=["parent_id"])
            .set_index("parent_id")
            .groupby("parent_id")
            .agg(agg)
        )
        df = df.groupby("parent_id").sum()
        df.index.name = id_name

        if agg == "sum":
            df = df[df.sum(1) > 0]
            df = df[~df.index.isin(data.index)]
        else:
            df = df.dropna()

        parents[parents["parent_id"].isin(df.index)] = pd.NA
        if len(df) == 0:
            break

        data = pd.concat([data, df], verify_integrity=True)
    else:
        raise RuntimeError("maximum number of iterations")

    return data


def _aggregate_secondary(data, agg="sum"):
    """
    Aggregate children of secondary parents relations.
    """
    id_name = data.index.name

    parents = db.mundi["alt_parents"].dropna()
    parents = parents["alt_parents"].str[1:].str.split(";").reset_index()

    parents = pd.DataFrame(
        list((idx, p) for _, (idx, parents) in parents.iterrows() for p in parents),
        columns=[id_name, "parent_id"],
    )

    return (
        pd.merge(parents, data.reset_index(), on=id_name)
        .drop(columns=id_name)
        .set_index("parent_id")
        .groupby("parent_id")
        .agg(agg)
    )


def child_parent_m2m(data, names=["child_id", "parent_id"], primary=True, secondary=True):
    """
    Return a m2m table with (child_id, parent_id) pairs.
    """
    out = []
    if primary:
        out.append(
            data.mundi["parent_id"]
            .dropna()
            .reset_index()
            .rename({"id": "child_id"}, axis=1)
        )

    if secondary:
        parents = (
            db.mundi["alt_parents"]["alt_parents"]
            .dropna()
            .reset_index()
            .rename({"alt_parents": "parent_id", "id": "child_id"}, axis=1)
        )
        out.append(
            pd.DataFrame(
                list(
                    (idx, p)
                    for _, (idx, parents) in parents.iterrows()
                    for p in parents[1:].split(";")
                ),
                columns=["child_id", "parent_id"],
            )
        )

    if not out:
        raise ValueError("must provide primary or secondary children")
    elif len(out) == 1:
        return out[0]
    else:
        return pd.concat(out).drop_duplicates().reset_index(drop=True)



import importlib
import time
from pathlib import Path

import click


@click.group()
def plugin():
    pass


@plugin.command(name="prepare")
@click.argument("package")
@click.option("--silent", "-s", is_flag=True, help="Do not print debug messages")
def prepare_cmd(**kwargs):
    """
    Prepare data for mundi plugin.
    """
    prepare(**kwargs)


def prepare(package, silent=False):
    """
    Prepare data for mundi plugin.
    """

    from .plugin import utils

    log = lambda *args: None if silent else click.echo(*args)
    t0 = time.time()

    log("Cleaning files...")
    utils.clean_processed_data(package, verbose=not silent)
    log("")

    utils.execute_prepare_scripts(package, verbose=not silent)
    log(f"Finished in {time.time() - t0:.2} seconds.")


@plugin.command(name="compile")
@click.argument("package")
@click.argument("kind")
@click.option("--silent", "-s", is_flag=True, help="Do not print debug messages")
@click.option("--output", "-o", help="Name of output file")
@click.option(
    "--fix", "-f", multiple=True, help="Name of functions used to post-process data"
)
def compile_cmd(**kwargs):
    """
    Compile prepared data for mundi plugin.
    """
    compile(**kwargs)


def compile(package, kind=None, output=None, fix=(), silent=False):
    """
    Compile prepared data for mundi plugin.
    """
    from .plugin import utils

    log = lambda *args: None if silent else click.echo(*args)
    t0 = time.time()

    log(f"Collecting processed files for {package}/{kind}")
    data = utils.collect_processed_data(package, kind)
    try:
        mod = importlib.import_module(package)
    except ImportError as e:
        raise SystemExit(e)

    # Apply fixes
    for name in fix:
        if callable(name):
            fn = name
            log(f"Applying fix: {fn.__name__}...")
            data = fn(package, kind, data)
            continue

        log(f"Applying fix: {name}...")

        # Normalize name
        if "." not in name:
            name = f"mundi.fix.{name}"
        elif name.startswith("."):
            name = package + name

        # Extract optional arguments
        name, _, args = name.partition("(")
        if args:
            args = [args.rstrip(")").split(",")]
        else:
            args = ()

        # Import function and apply
        pre, _, post = name.rpartition(".")
        try:
            fix_mod = importlib.import_module(pre)
            fn = getattr(fix_mod, post)
        except (ImportError, AttributeError):
            raise SystemExit(f"invalid python name: {name}")

        data = fn(package, kind, data, *args)

    # Save output
    if output is None:
        print(data)
    else:
        path = Path(mod.__file__).parent.resolve() / "databases" / output
        utils.save_path(data, path, kind)
        log(f" - Database saved to {path}")

    log(f"Finished in {time.time() - t0:.2} seconds.")



#
# This script creates the sus-cities.csv and sus-macros.csv scripts
#

import json
from collections import defaultdict
from pathlib import Path

import pandas as pd

import mundi

COLUMNS = mundi.DATA_COLUMNS["mundi"]
PATH = Path(__file__).parent
STATE_CODES = json.load(open(PATH / "state_codes.json"))
CITY_CODES = json.load(open(PATH / "city-codes.json"))

df = pd.read_csv(PATH / "regionais-de-saude-macro.csv", index_col=0).drop(
    columns=["macro_id"]
)
df["region_id"] = df["region_id"].apply(lambda x: f"BR-SUS:{x}").astype("string")
df["city_id"] = df["city_id"].apply(lambda x: f"BR-{CITY_CODES[str(x)]}").astype("string")
df["uf_id"] = df["region_id"].str[7:9]
df["state_id"] = "BR-" + df["uf"]


#
# SUS macro healthcare regions
#
macro_ids = defaultdict(dict)
for _, (uf, uf_id, name) in df[["uf", "uf_id", "macro_name"]].iterrows():
    db = macro_ids[uf]
    if name not in db:
        db[name] = f"BR-SUS:{uf_id}{len(db) + 1:02};BR-{uf}"

macros = (
    pd.concat(pd.Series(rs) for rs in macro_ids.values())
    .reset_index()
    .astype("string")
    .rename({"index": "name"}, axis=1)
)
macros = pd.concat(
    [
        macros,
        macros.pop(0)
        .str.partition(";")
        .drop(columns=1)
        .rename({0: "id", 2: "state_id"}, axis=1),
    ],
    axis=1,
).set_index("id")


#
# Merge macro-regions with cities
#
cities = (
    pd.merge(
        df[["state_id", "macro_name", "city_id"]].rename(
            {"macro_name": "name", "city_id": "id"}, axis=1
        ),
        macros.reset_index().rename({"id": "sus_id"}, axis=1),
        on=["state_id", "name"],
    )
    .drop(columns="state_id")
    .set_index("id")
)

macros = macros.rename({"state_id": "parent_id"}, axis=1)

print(macros)

macros.to_csv(PATH / "sus-macros.csv")
cities.to_csv(PATH / "sus-cities.csv")



import pytest

import mundi


class TestDataFrameAccessor:
    @pytest.fixture(scope="class")
    def db(self):
        db = mundi.countries()
        return db.iloc[:10]

    def test_extract_extra_columns(self, db):
        extra = db.mundi["income_group"]
        assert extra.shape == (10, 1)

    def test_extract_and_append_extra_columns(self, db):
        extra = db.mundi[..., "income_group"]
        assert extra.shape == (10, 2)

    def test_extract_multiple_columns(self, db):
        extra = db.mundi["region", "income_group"]
        assert extra.shape == (10, 2)



import pycountry
import pandas as pd
from pathlib import Path
import numpy as np
import mundi

COLUMNS = mundi.DATA_COLUMNS["mundi"]
COL_RENAME = {"code": "id", "parent_code": "parent_id"}
BLACKLIST = {"BR-FN"}
PATH = Path(__file__).parent
SUBDIVISION_TYPES = {
    # 'prefecture', 'indigenous region', 'administrative region', 'autonomous
    # district', 'borough', 'london borough', 'canton', 'department', 'oblast',
    # 'republic', 'overseas territorial collectivity', 'geographical unit',
    # 'metropolitan region', 'municipalities', 'popularates', 'entity', 'metropolitan
    # district', 'geographical region', 'autonomous territorial
    # unit', 'geographical entity', 'autonomous republic', 'district', 'metropolitan
    # department', 'council area', 'parish', 'development region', 'chains (of
    # islands)', 'economic prefecture', 'territory', 'unitary authority', 'special
    # island authority', 'overseas region', 'administrative territory',
    # 'capital territory', 'capital district', 'union territory', 'special zone',
    # 'self-governed part', 'constitutional province', 'administration', 'autonomous
    # sector', 'nation', 'zone', 'autonomous community',
    # 'quarter', 'local council', 'capital metropolitan city', 'province',
    #  'overseas department',  'island group',
    # 'commune', 'state', 'territorial unit', 'two-tier county',
    # 'special district', 'division', 'area', 'administrative atoll',
    # 'autonomous region', 'federal territories', 'autonomous
    # province', 'dependency', 'emirate',  'arctic region', 'rayon', 'regional
    # council', 'island', 'town council', 'metropolitan cities',
    # 'island council', 'county', 'special region', 'outlying area', 'governorate',
    # 'region', 'special administrative region',  'federal
    # dependency'
    #
    # Primitive types
    # 'country'
    #
    # Derived types with corresponding sub-type
    # State-like
    "state": ("state", "federal district"),
    # City-like
    "city": ("city", pd.NA),
    "city with county rights": ("city", "county rights"),
    "republican city": ("city", "republican city"),
    "special city": ("city", "special"),
    "capital city": ("city", "capital"),
    "city corporation": ("city", "city corporation"),
    "autonomous city": ("city", "autonomous"),
    "municipality": ("city", "municipality"),
    "autonomous municipality": ("city", "autonomous municipality"),
    "special municipality": ("city", "special municipality"),
}

#
# This script uses pycountry (https://pypi.org/project/pycountry/) as the source
# for information about countries and the main sub-divisions.
#
df = pd.DataFrame([s._fields for s in pycountry.subdivisions if s.code not in BLACKLIST])
df = df.rename(COL_RENAME, axis=1).set_index("id").drop(columns=["parent"])

df["short_code"] = pd.Series(df.index).astype("string").str.partition("-")[2].values
df["short_code"] = codes = df["short_code"].astype("string")
df["numeric_code"] = np.where(codes.str.isdigit(), codes, pd.NA)
df["long_code"] = df["alt_parents"] = pd.NA

# Fix region types/subtypes
data = df["type"].apply(lambda x: SUBDIVISION_TYPES.get(x.lower(), (x.lower(), pd.NA)))
types_df = pd.DataFrame(list(data), columns=["type", "subtype"], index=df.index)
df["type"] = types_df["type"]
df["subtype"] = types_df["subtype"]

df = df.astype("string")[COLUMNS]
assert len(set(df.index)) == len(df)

path = PATH / "processed" / "mundi-A1-subdivisions.pkl"
df.to_pickle(path)
print(f"Subdivisions data saved to {path}")



from .region import REGION_PLUGINS, Region
from ..db import db


def region_mundi_plugin(db, attr):
    """
    Function that extracts attributes from db for the given region.
    """

    def plugin(region: Region):
        cur = db.raw_sql(f'SELECT id, {attr} FROM {{table}} WHERE id = "{region.id}";')
        try:
            (_, value) = next(iter(cur))
        except IndexError:
            return None
        return value

    plugin.__name__ = attr
    plugin.__qualname__ = attr
    return plugin


REGION_PLUGINS.update(
    name=region_mundi_plugin(db, "name"),
    type=region_mundi_plugin(db, "type"),
    subtype=region_mundi_plugin(db, "subtype"),
    short_code=region_mundi_plugin(db, "short_code"),
    long_code=region_mundi_plugin(db, "long_code"),
    numeric_code=region_mundi_plugin(db, "numeric_code"),
    country_code=region_mundi_plugin(db, "country_code"),
    parent_id=region_mundi_plugin(db, "parent_id"),
    alt_parents=region_mundi_plugin(db, "alt_parents"),
)



from pandas import CategoricalDtype

from . import loader

REGIONS = {
    "north-america": "North America",
    "south-asia": "South Asia",
    "sub-saharan-africa": "Sub-Saharan Africa",
    "europe": "Europe & Central Asia",
    "latin-america": "Latin America & Caribbean",
    "middle-east": "Middle East & North Africa",
    "east-asia": "East Asia & Pacific",
}
INCOME_GROUPS = {
    "low": "Low income",
    "lower-middle": "Lower middle income",
    "upper-middle": "Upper middle income",
    "high": "High income",
}

IncomeGroup = CategoricalDtype(categories=INCOME_GROUPS, ordered=True)
Region = CategoricalDtype(categories=REGIONS, ordered=False)


@loader.filtering_from_data(["region"])
def load_region():
    return loader.load_database("un.pkl.gz").astype(Region)


@loader.filtering_from_data(["income_group"])
def load_income_group():
    return loader.load_database("un.pkl.gz").astype(IncomeGroup)



import sqlite3
from collections.abc import Sequence
from functools import lru_cache
from pathlib import Path

import pandas as pd

from .constants import DATA_COLUMNS

SQLITE_PATH = Path(__file__).parent.resolve() / "databases" / "db.sqlite"
COLUMNS = DATA_COLUMNS["mundi"]
UN_COLUMNS = ["region", "income_group"]
UN_COLUMN_TYPES = {"region": "category", "income_group": "category"}


class RegionsDB:
    """
    Implements the countries(), regions() and region() callables.
    """

    columns = COLUMNS
    column_types = {k: "string" for k in columns}
    column_types["type"] = "category"

    def __init__(self, ref, path=SQLITE_PATH, **kwargs):
        self._path = path
        self._ref = ref
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __call__(self, *args, **kwargs):
        kws = {k: v for k, v in kwargs.items() if k in self.columns}
        if kws:
            kwargs = {k: v for k, v in kwargs.items() if k not in kws}
        df = self.query(**kws)
        if args:
            df = df.mundi.extend(*args)
        if kwargs:
            df = df.mundi.select(**kwargs)
        return df

    def __hash__(self):
        return id(self)

    def query(self, case_sensitive=False, cols=("id", "name"), **kwargs):
        """
        Query database with given filtering parameters.
        """

        cols = "*" if cols is None else ", ".join(cols)
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        filters = [f"{filter_command(k, v, case_sensitive)}" for k, v in kwargs.items()]
        filters_suffix = ""
        if filters:
            filters_suffix = " WHERE " + " AND ".join(filters)
        cmd = "SELECT %s FROM {table}%s;" % (cols, filters_suffix)
        return self.sql(cmd, index=cols != "id")

    def get(self, *args, **kwargs):
        """
        Get single element from database.
        """

        if args:
            (id_,) = args
        else:
            id_ = None

        if kwargs and id_:
            raise TypeError("canoot pass id and query parameters simultaneously")
        elif id_:
            cmd = 'SELECT * FROM {table} WHERE id="%s" COLLATE NOCASE;' % id_
            data = self.sql(cmd)
        elif kwargs:
            data = self.query(**kwargs)
        else:
            raise TypeError("must pass id or some filter for query parameters")

        if len(data) == 0:
            raise LookupError("no element found with the given ID")
        elif len(data) > 1:
            raise LookupError("found multiple elements")

        res = data.iloc[0]
        res.name = res.get("id", data.index[0])
        return res

    def column_loader(self, col):
        """
        Return a loader function for the given column.
        """

        return lambda df: self.load_column(col, df.index)

    def load_column(self, col, ids=None):
        """
        Load column from database.
        """

        ids = ",".join(map(repr, ids))
        cmd = "SELECT id, %s FROM {table} WHERE id IN (%s);" % (col, ids)
        return self.sql(cmd, index=True)

    def sql(self, sql, copy=True, index=False):
        """
        Execute raw SQL command.
        """

        sql = sql.format(table=self._ref)
        df = read_sql(self, self._path, sql, index=index)
        return df.copy() if copy else df

    def raw_sql(self, sql):
        """
        Execute raw SQL command.
        """

        sql = sql.format(table=self._ref)
        with sqlite3.connect(self._path) as conn:
            c = conn.cursor()
            return c.execute(sql)


def filter_command(k, v, case_sensitive=True):
    if isinstance(v, str) or not isinstance(v, Sequence):
        cmd = f'{k} = "{repr(v)[1:-1]}"'
    else:
        sep = " or "
        args = (filter_command(k, vi) for vi in v)
        cmd = f"({sep.join(args)})"

    if not case_sensitive:
        cmd += " COLLATE NOCASE"
    return cmd


@lru_cache(256)
def read_sql(db, path, sql, index=True):
    kwargs = {"index_col": "id"} if index else {}

    with sqlite3.connect(path) as conn:
        df = pd.read_sql(sql, conn, **kwargs)

        if index:
            cols = [c for c in db.columns if c in df.columns]
            types = {c: db.column_types[c] for c in cols}
            df = df[cols].astype(types)

    return df


db = RegionsDB("mundi")
db_un = RegionsDB("un", columns=UN_COLUMNS, column_types=UN_COLUMN_TYPES)



from invoke import task


@task
def test(ctx):
    ctx.run("pytest tests/ --cov")
    ctx.run("black --check .")
    ctx.run("pycodestyle")


@task
def prepare_data(ctx, fast=False):
    if not fast:
        ctx.run("rm mundi/databases/* -rfv")
        ctx.run("python -m mundi prepare mundi")
        print()

    ctx.run("python -m mundi compile mundi mundi -o db.sqlite -f fix_types")
    print()
    ctx.run("python -m mundi compile mundi un -o db.sqlite -f fix_types")
    print()
    ctx.run("python -m mundi compile mundi un -o un.pkl.gz")



import re
from functools import lru_cache
from typing import Union

import numpy as np
import pandas as pd
import sidekick as sk

from .types import Region

db = sk.import_later(".db:db", package=__package__)
ISO2 = re.compile(r"[A-Z]{2}")
ISO3 = re.compile(r"[A-Z]{3}")
MUNDI_CODE = re.compile(r"[A-Z]{2}-\w+(:\w+)*")
TYPES_HIERARCHY = ["state", "city", "district", "region"]


def regions(country=None, **kwargs) -> pd.DataFrame:
    """
    Query the regions/sub-divisions database.
    """
    if country and "country_code" in kwargs:
        raise TypeError("cannot specify country and country_code")
    elif country:
        kwargs["country_code"] = country_code(country)
    return db.query(**kwargs)


def countries(**kwargs) -> pd.DataFrame:
    """
    Query the country database.
    """
    return regions(type="country", **kwargs)


def region(*args, country=None, **kwargs) -> Region:
    """
    Query the regions/sub-divisions database.
    """
    if country:
        kwargs["country_code"] = country_code(country)

    if args:
        (ref,) = args
        if isinstance(ref, Region):
            return ref
        row = db.get(code(ref))
    else:
        row = db.get(**kwargs)
    return Region(row.name)


@lru_cache(1024)
def country_code(code: str) -> str:
    """
    Return the country code for the given country.

    Similar to the code() function, but only accept valid countries.
    """
    if isinstance(code, Region):
        if code.type != "country":
            raise ValueError(f"region is not a country: {code}")
        return code.id

    if ISO2.fullmatch(code.upper()):
        try:
            db.get(code)
            return code.upper()
        except LookupError:
            pass

    elif ISO3.fullmatch(code.upper()):
        try:
            res = db.get(long_code=code.upper(), type="country")
            return res.name
        except LookupError:
            pass

    if code.isdigit():
        res = db.get(numeric_code=code, type="country")
        return res.name

    elif "/" not in code:
        res = db.get(name=code, type="country")
        return res.name

    raise LookupError(code)


@lru_cache(32_000)
def code(code: Union[Region, str]) -> str:
    """
    Return the mundi code for the given region.
    """
    if isinstance(code, Region):
        return code.id

    try:
        return country_code(code)
    except LookupError:
        pass

    if MUNDI_CODE.fullmatch(code.upper()):
        try:
            res = db.get(code)
            return res.name
        except LookupError:
            pass
        country, _, division = code.partition("-")

    elif "/" in code:
        country, _, division = code.partition("/")

    else:
        raise LookupError(code)

    country = country_code(country)
    return _subdivision_code(country, division)


@lru_cache(32_000)
def _subdivision_code(country: str, subdivision: str) -> str:
    """
    Return the mundi code for the given subdivision of a country.
    """
    if subdivision.isdigit():
        res = db.get(numeric_code=subdivision, country_code=country)
        return f"{country}-{res.name}"

    else:
        for lookup in ["short_code", "long_code"]:
            kwargs = {lookup: subdivision, "country_code": country}
            try:
                res = db.get(**kwargs)
                return f"{country}-{res.name}"
            except LookupError:
                pass

        values = db.query(
            country_code=country, name=subdivision, cols=("id", "type", "subtype")
        )

        if len(values) == 1:
            return values.index[0]
        elif len(values) > 1:
            pos = np.argsort([TYPES_HIERARCHY.index(x) for x in values["type"]])
            return values.index[pos[0]]
        else:
            raise LookupError(code)



REGION_DESCRIPTIONS = {
    "north-america": "North America",
    "south-asia": "South Asia",
    "sub-saharan-africa": "Sub-Saharan Africa",
    "europe": "Europe & Central Asia",
    "latin-america": "Latin America & Caribbean",
    "middle-east": "Middle East & North Africa",
    "east-asia": "East Asia & Pacific",
}
INCOME_GROUP_DESCRIPTIONS = {
    "low": "Low income",
    "lower-middle": "Lower middle income",
    "upper-middle": "Upper middle income",
    "high": "High income",
}
DATA_COLUMNS = {
    "mundi": [
        "name",
        "type",
        "subtype",
        "short_code",
        "numeric_code",
        "long_code",
        "country_code",
        "parent_id",
        "alt_parents",
    ],
    "un": ["region", "income_group"],
}
DATA_TYPES = {
    "name": "string",
    "type": "category",
    "subtype": "category",
    "short_code": "string",
    "numeric_code": "string",
    "long_code": "string",
    "country_code": "string",
    "parent_id": "string",
    "alt_parents": "string",
    "region": "category",
    "income_group": "category",
}



import json
from collections import namedtuple
from functools import lru_cache
from pathlib import Path

import pandas as pd

import mundi

COLUMNS = mundi.DATA_COLUMNS["mundi"]
PATH = Path(__file__).parent
STATE_CODES = json.load(open(PATH / "state_codes.json"))


@lru_cache(1)
def territory():
    """
    Full data from IBGE about the division of territories up to the level of
    districts.
    """

    # Read raw data from IBGE
    out = pd.read_csv(PATH / "divisao-do-territorio-distritos.csv")
    out.columns = [
        "state",
        "state_name",
        "meso_code",
        "meso_name",
        "micro_code",
        "micro_name",
        "city_short_code",
        "city_code",
        "city_name",
        "district_short_code",
        "district_code",
        "district_name",
    ]
    out["state_code"] = out["state"]
    out["state"] = out["state_name"].apply(STATE_CODES.get)
    return out


@lru_cache(1)
def macro_regions():
    """
    Macro regions, i.e., Norte, Nordeste, Sudeste, Sul e Centro-Oeste.
    """

    out = pd.DataFrame(
        [
            ["Norte", "1"],
            ["Nordeste", "2"],
            ["Sudeste", "3"],
            ["Sul", "5"],
            ["Centro-Oeste", "5"],
        ],
        index=[f"BR-{i}" for i in range(1, 6)],
        columns=["name", "short_code"],
    )
    out["type"] = "region"
    out["subtype"] = "macro-region"
    out["numeric_code"] = out["short_code"]
    out["country_code"] = "BR"
    out["parent_id"] = "BR"
    out["alt_parents"] = out["long_code"] = pd.NA
    return out.astype("string")[COLUMNS]


@lru_cache(1)
def states():
    """
    Dataframe with all Brazillian states.
    """
    # Load states from prepared pycountry database
    out = (
        pd.Series(STATE_CODES, name="short_code")
        .reset_index()
        .astype("string")
        .rename({"index": "name"}, axis=1)
    )
    out.index = "BR-" + out["short_code"]

    # Save numeric codes for states
    codes = territory()[["state", "state_code"]].drop_duplicates()
    codes.index = "BR-" + codes.pop("state").astype("string")

    # Assign region numbers
    out["numeric_code"] = codes["state_code"].apply(str).astype("string")
    out["parent_id"] = "BR-" + out["numeric_code"].str[:1]
    out["type"] = "state"
    out["country_code"] = "BR"
    out["alt_parents"] = out["subtype"] = out["long_code"] = pd.NA

    out.loc["BR-DF", "subtype"] = "federal district"
    return out.astype("string")[COLUMNS]


@lru_cache(1)
def SUS():
    """
    SUS healthcare regions.
    """

    out = pd.read_csv(PATH / "sus-macros.csv", dtype="string")
    out["type"] = "region"
    out["subtype"] = "healthcare region"
    out["country_code"] = "BR"
    out["short_code"] = out["id"].str[7:]
    out["numeric_code"] = out["id"].str[7:]
    out["long_code"] = out["id"].str[3:]
    out["alt_parents"] = pd.NA
    out = out.set_index("id")
    return out.astype("string")


@lru_cache(1)
def subdivisions():
    """
    Subdivisions

    Include SUS regions as parent_extra after loading raw _subdivisions
    """
    out = _subdivisions()
    sus = pd.read_csv(PATH / "sus-cities.csv", index_col=0, dtype="string")
    out["alt_parents"] = ";" + sus["sus_id"]
    return out


def _subdivisions():
    """
    Subdivisions.

    Do not fix SUS healthcare regions.
    """
    # Collect dataframes
    cols = ["short_code", "name", "type", "subtype", "parent_id", "long_code"]
    Row = namedtuple("Row", cols)

    # Regions
    meso = set()
    micro = set()
    city = set()
    district = set()

    for _, row in territory().iterrows():
        meso_id = f"{row.state_code}{row.meso_code:02}"
        micro_id = f"{meso_id}{row.micro_code:02}"
        city_long_id = str(row.city_code)
        city_id = str(row.city_code)[:-1]  # strip unused last digit
        district_id = str(row.district_code)

        meso.add(
            Row(meso_id, row.meso_name, "region", "meso-region", str(row.state), pd.NA)
        )
        micro.add(Row(micro_id, row.micro_name, "region", "micro-region", meso_id, pd.NA))
        city.add(
            Row(city_id, row.city_name, "city", "municipality", micro_id, city_long_id)
        )
        district.add(
            Row(district_id, row.district_name, "district", pd.NA, city_id, pd.NA)
        )

    def mk_table(data):
        df = pd.DataFrame(list(data), columns=cols)
        df["numeric_code"] = df["long_code"].fillna(df["short_code"])
        df["parent_id"] = "BR-" + df["parent_id"]
        df["country_code"] = "BR"
        df["alt_parents"] = pd.NA

        df.index = "BR-" + df["numeric_code"]
        df.index.name = "id"
        return df.astype("string")

    br_meso = mk_table(meso)
    br_micro = mk_table(micro)
    br_city = mk_table(city)
    br_district = mk_table(district)

    df = pd.concat([br_meso, br_micro, br_city, br_district])
    df["type"] = df["type"]

    return df.astype("string")[COLUMNS]


# Save tables
for i, fn in enumerate([macro_regions, states, SUS, subdivisions], start=1):
    name = fn.__name__.rstrip("s")
    path = PATH / "processed" / f"mundi-C{i}-{name}.pkl"
    df = fn()
    assert len(set(df.index)) == len(df)
    df.to_pickle(path)
    print(f"BR data saved to {path}")



from typing import Optional
from weakref import WeakValueDictionary

import pandas as pd

from ..db import db

REGION_PLUGINS = {}
REGION_DB = WeakValueDictionary()


class Region:
    """
    Represent entries/rows in the mundi db.
    """

    name: str

    def __new__(cls, id):
        if not isinstance(id, str):
            raise TypeError(f"expect string, got {id.__class__.__name__}")
        try:
            return REGION_DB[id]
        except KeyError:
            new = object.__new__(cls)
            new.__dict__["id"] = id
            REGION_DB[id] = new
            return new

    def __hash__(self):
        return hash(self.id)

    def __getstate__(self):
        return self.id

    def __setstate__(self, id):
        self.__dict__["id"] = id

    def __getitem__(self, key):
        return self._get_field(key)

    def __setattr__(self, attr, value):
        raise AttributeError("Region objects are immutable")

    def __getattr__(self, item):
        if item.startswith("_"):
            raise AttributeError(item)
        try:
            return self._get_field(item)
        except KeyError:
            raise AttributeError(item)

    def __repr__(self):
        try:
            args = f", name={self.name!r}"
        except AttributeError:
            args = ""
        return f"{self.__class__.__name__}({self.id!r}{args})"

    def __str__(self):
        return (
            f"Mundi Region\n"
            f"id       : {self.id}\n"
            f"name     : {self.name}\n"
            f"type     : {self.type}\n"
            f"subtype  : {self.subtype}"
        )

    def __lt__(self, other):
        if isinstance(other, Region):
            return self.id < other.id
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, Region):
            return self.id > other.id
        return NotImplemented

    def _get_field(self, key):
        fget = REGION_PLUGINS[key]
        self.__dict__[key] = value = fget(self)
        return value

    #
    # Conversions
    #
    def to_dict(self, *fields: str) -> dict:
        """
        Convert Region to a dictionary including all requested attributes.
        """
        get = self._get_field
        return {"id": self.id, "name": self.name, **{f: get(f) for f in fields}}

    def to_series(self, *fields: str) -> pd.Series:
        """
        Convert Region to a pandas Series element including the requested
        attributes.
        """
        return pd.Series(self.to_dict(*fields))

    #
    # Hierarchies
    #
    @property
    def parent(self) -> Optional["Region"]:
        pk = self.parent_id
        return Region(pk) if pk else None

    def children(self, dataframe=False, deep=False, which="both"):
        """
        Return list of children.
        """
        if dataframe:
            raise NotImplementedError

        primary = which in ("primary", "both")
        secondary = which in ("secondary", "both")
        ids = self._children_ids(primary, secondary)
        if deep:
            non_processed = ids.copy()
            while non_processed:
                id = non_processed.pop()
                new = Region(id)._children_ids(primary, secondary)
                non_processed.update(new - ids)
                ids.update(new)

        return [Region(id) for id in ids]

    def _children_ids(self, primary, secondary) -> set:
        ids = set()

        if primary:
            res = db.query(parent_id=self.id, cols=("id",))
            ids.update(res["id"])

        if secondary:
            cmd = f"SELECT id FROM mundi WHERE alt_parents LIKE '%;{self.id}%';"
            ids.update(map(lambda x: x[0], db.raw_sql(cmd)))

        return ids

    def parents(self, dataframe=False):
        """
        Return list of parents.
        """
        raise NotImplementedError



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['mundi', 'mundi.plugin', 'mundi.types']

package_data = \
{'': ['*'], 'mundi': ['databases/*']}

install_requires = \
['pandas', 'sidekick']

extras_require = \
{'dev': ['black', 'twine', 'pycountry', 'coverage~=5.0', 'pytest~=5.0']}

setup(name='mundi',
      version='0.2.1',
      description='Centralize data from countries and regions around the World and expose it',
      author='Fábio Macêdo Mendes',
      author_email='fabiomacedomendes@gmail.com',
      url='https://github.com/pydemic/mundi',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )



from abc import ABC

PLUGIN_DB = {}


class Plugin(ABC):
    """
    Plugins make data available for mundi.
    """

    def __init__(self, name):
        self.name = name

    def register(self, force=False):
        """
        Register plugin in the plugin database.
        """


class ColumnPlugin(Plugin, ABC):
    """
    pass
    """



"""
Centralize data from countries and regions around the World and expose it
as Pandas dataframes.

Mundi has a plug-in system that expands the data sets to many other
specialized uses.
"""
__author__ = "Fábio Mendes"
__version__ = "0.2.1"

from . import pandas as _pandas_mod
from .functions import regions, countries, region, country_code, code
from .types import Region
from .constants import *



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

import mundi
from mundi import Region


class TestDb:
    @pytest.fixture
    def country(self):
        return mundi.region("br")

    @pytest.fixture
    def countries(self):
        return mundi.countries(["br", "it", "cn"])

    def test_load_countries(self):
        db = mundi.countries()
        assert len(db) == 255

    def test_load_countries_is_an_alias_to_regions(self):
        db1 = mundi.functions.regions(type="country")
        db2 = mundi.functions.countries()
        assert_frame_equal(db1, db2)

    def test_load_states_from_regions(self):
        db = mundi.functions.regions(type="state", country="BR")
        print(db)
        assert len(db) == 27

    def test_load_country(self):
        br = mundi.region("BR")
        assert isinstance(br, Region)
        assert br.id == "BR"
        assert br["name"] == "Brazil"
        assert br["type"] == "country"
        assert br["short_code"] == "BR"
        assert br["numeric_code"] == "076"
        assert br["long_code"] == "BRA"
        assert br["country_code"] is None
        assert br["parent_id"] == "XSA"

        ar = mundi.region("ar")
        assert isinstance(ar, Region)
        assert ar["name"] == "Argentina"
        assert ar["type"] == "country"
        assert ar["short_code"] == "AR"
        assert ar["numeric_code"] == "032"
        assert ar["long_code"] == "ARG"
        assert ar["country_code"] is None
        assert br["parent_id"] == "XSA"


class TestBR:
    def test_fn_is_not_a_state(self):
        with pytest.raises(LookupError):
            mundi.region("BR-FN")

    def test_df_is_a_state(self):
        with pytest.raises(LookupError):
            mundi.region("BR-FN")



from pathlib import Path

import pandas as pd
import pycountry

import mundi

inv = lambda d: {v: k for k, v in d.items()}
PATH = Path(__file__).parent.resolve()
COL_RENAME = {"alpha_2": "short_code", "alpha_3": "long_code", "numeric": "numeric_code"}
REGION_MAP = inv(mundi.REGION_DESCRIPTIONS)
INCOME_MAP = inv(mundi.INCOME_GROUP_DESCRIPTIONS)

#
# This script uses pycountry (https://pypi.org/project/pycountry/) as the main
# source of information about countries and sub-divisions.
#
df = pd.DataFrame([c._fields for c in pycountry.countries])
df = (
    df.rename(COL_RENAME, axis=1)
    .drop(columns=["official_name", "common_name"])
    .astype("string")
    .set_index("short_code", drop=False)
)
df.index.name = "id"

df["type"] = "country"
df["subtype"] = pd.NA
df["country_code"] = pd.NA
df = df.astype("string")

#
# Load UN classifications from world bank data
#
un_classes = pd.read_csv(PATH / "world-bank-summary.csv", index_col=0, dtype="string")
un_classes = un_classes[["region", "income_group"]]
un_classes["id"] = df["long_code"].reset_index().set_index("long_code")
un_classes = un_classes.dropna().set_index("id")

# Save to dataframe
un_classes["region"] = un_classes["region"].apply(REGION_MAP.get)
un_classes["income_group"] = un_classes["income_group"].apply(INCOME_MAP.get)

path = PATH / "processed" / "un.pkl"
un_classes.astype("category").to_pickle(path)
print(f"UN data saved to {path}")


#
# Loading continents
#

# Load continent codes
# https://en.wikipedia.org/wiki/List_of_sovereign_states_and_dependent_territories_by_continent_(data_file)#Data_file
data = pd.read_csv(PATH / "countries-to-continents.csv", index_col=0, dtype="string")
df["parent_id"] = "X" + data["continent_id"]

# Load continents
data = pd.read_csv(PATH / "continents.csv").fillna("NA").set_index("id")
data["type"] = "continent"
data["parent_id"] = "XX"
data["short_code"] = data.index
df = pd.concat([df, data]).astype("string")

# Transcontinental countries are assigned to their secondary continent in the
# alt_parent column.
# https://en.wikipedia.org/wiki/List_of_transcontinental_countries
df["alt_parents"] = pd.Series(
    {
        "AZ": "XEU",
        "AM": "XAS",
        "CY": "XAS",
        "GE": "XAS",
        "KZ": "XEU",
        "UM": "XOC",
        "RU": "XAS",
        "TR": "XEU",
        "EG": "XAS",
    },
    dtype="string",
)
df["alt_parents"] = ";" + df["alt_parents"]

#
# Saving results
#
df = df.astype("string")[mundi.DATA_COLUMNS["mundi"]]
assert len(set(df.index)) == len(df)

path = PATH / "processed" / "mundi-A1-countries.pkl"
df.to_pickle(path)
print(f"Country data saved to {path}")
